# # for testing
from njab.sklearn import scoring
import numpy as np


def test_get_label_binary_classification():
    for y_true, y_pred, label in zip(np.array([1, 1, 0, 0]),
                                     np.array([1, 0, 1, 0]),
                                     ['TP', 'FN', 'FP', 'TN']):
        assert scoring.get_label_binary_classification(y_true, y_pred) == label



"""Analysis of covariance using pingouin and statsmodels."""
from __future__ import annotations

import numpy as np
import pandas as pd
import pingouin as pg
import statsmodels


def ancova_pg(
    df_long: pd.DataFrame,
    feat_col: str,
    dv: str,
    between: str,
    covar: list[str] | str,
) -> pd.DataFrame:
    """ Analysis of covariance (ANCOVA) using pg.ancova
    https://pingouin-stats.org/generated/pingouin.ancova.html

    Adds multiple hypothesis testing correction by Benjamini-Hochberg
    (qvalue, rejected)

    Parameters
    ----------
    df_long : pd.DataFrame
        should be long data format
    feat_col : str
        feature column (or index) name
    dv : str
        Name of column containing the dependant variable, passed to pg.ancova
    between : str
        Name of column containing the between factor, passed to pg.ancova
    covar : list, str
        Name(s) of column(s) containing the covariate, passed to pg.ancova
    fdr : float, optional
        FDR treshold to apply, by default 0.05

    Returns
    -------
    pd.DataFrame
        Columns:  [ 'Source',
                    'SS',
                    'DF',
                    'F',
                    'p-unc',
                    'np2',
                    '{feat_col}',
                    '-Log10 pvalue',
                    'qvalue',
                    'rejected']
    """
    scores = []
    # num_covar = len(covar)

    for feat_name, data_feat in df_long.groupby(feat_col):
        # ? drop duplicated colummns in long data format?
        ancova = pg.ancova(data=data_feat, dv=dv, between=between, covar=covar)
        ancova[feat_col] = feat_name
        scores.append(ancova)
    scores = pd.concat(scores)
    scores['-Log10 pvalue'] = -np.log10(scores['p-unc'])

    return scores


def add_fdr_scores(scores: pd.DataFrame,
                   random_seed: int = None,
                   alpha: float = 0.05,
                   method: str = 'indep',
                   p_val_column: str = 'p-unc') -> pd.DataFrame:
    """Add FDR scores based on p-values in p_val_column."""
    if random_seed is not None:
        np.random.seed(random_seed)
    reject, qvalue = statsmodels.stats.multitest.fdrcorrection(
        scores[p_val_column], alpha=alpha, method=method)
    scores['qvalue'] = qvalue
    scores['rejected'] = reject
    return scores


class Ancova():
    """Base Ancova class."""

    def __init__(self,
                 df_proteomics: pd.DataFrame,
                 df_clinic: pd.DataFrame,
                 target: str,
                 covar: list[str],
                 value_name: str = 'intensity'):
        """
        Parameters
        ----------
        df_proteomics : pd.DataFrame
            proteomic measurements in wide format
        df_clinic : pd.DataFrame
            clinical data, containing `target` and `covar`
        target : str
            Variable for stratification contained in `df_clinic`
        covar : list[str]
            List of control varialbles contained in `df_clinic`
        value_name : str
            Name to be used for protemics measurements in long-format, default "intensity"
        """
        self.df_proteomics = df_proteomics
        self.df_clinic = df_clinic
        self.target = target
        self.covar = covar
        self.value_name = value_name

    def get_scores(self):
        data = (self.df_proteomics.loc[self.df_clinic[
            self.target].notna()].stack().to_frame(self.value_name).join(
                self.df_clinic))
        feat_col = data.index.names[-1]
        scores = ancova_pg(data,
                           feat_col=feat_col,
                           dv=self.value_name,
                           between=self.target,
                           covar=self.covar)
        return scores.set_index(feat_col)

    def ancova(self):
        raise NotImplementedError


def filter_residuals_from_scores(scores: pd.DataFrame,
                                 filter_for='Residual') -> pd.DataFrame:
    """Remove residual from pingouin ANCOVA list."""
    scores = scores[scores.Source != filter_for]
    return scores


class AncovaAll(Ancova):
    """Ancova with FDR on all variables except the constant
       of the linear regression for each.
    """

    def ancova(self, random_seed=123):

        scores = self.get_scores()
        scores = filter_residuals_from_scores(scores)
        # drop nan values (due to multicollinearity of features - i.e. duplicated features)
        scores = scores.dropna()
        scores = add_fdr_scores(scores, random_seed=random_seed)
        self.scores = scores
        return scores.set_index('Source', append=True)


def filter_all_covars_from_scores(scores: pd.DataFrame,
                                  filter_for: str) -> pd.DataFrame:
    """Only keep feature score from pingouin ANCOVA list."""
    scores = scores[scores.Source == filter_for]
    return scores


class AncovaOnlyTarget(Ancova):
    """Ancova with FDR on only the target variables p-values
    in the set of hypothesis."""

    def ancova(self, random_seed=123) -> pd.DataFrame:
        scores = self.get_scores()
        scores = filter_all_covars_from_scores(scores, filter_for=self.target)
        scores = add_fdr_scores(scores, random_seed=random_seed)
        self.scores = scores
        return scores.set_index('Source', append=True)



import logging
import typing

import omegaconf
import pandas as pd
import pandas.io.formats.format as pf

logger = logging.getLogger(__name__)


def set_pandas_options(max_columns: int = 100,
                       max_row: int = 30,
                       min_row: int = 20,
                       float_format='{:,.3f}') -> None:
    """Update default pandas options for better display."""
    pd.options.display.max_columns = max_columns
    pd.options.display.max_rows = max_row
    pd.options.display.min_rows = min_row
    set_pandas_number_formatting(float_format=float_format)


def set_pandas_number_formatting(float_format='{:,.3f}') -> None:
    pd.options.display.float_format = float_format.format
    # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.describe_option.html#pandas.describe_option
    pd.options.styler.format.thousands = ','
    # # https://github.com/pandas-dev/pandas/blob/main/pandas/io/formats/format.py#L1475
    # Originally found: https://stackoverflow.com/a/29663750/9684872

    try:
        base_class = pf.GenericArrayFormatter
    except AttributeError:
        base_class = pf._GenericArrayFormatter

    class IntArrayFormatter(base_class):

        def _format_strings(self):
            formatter = self.formatter or '{:,d}'.format
            fmt_values = [formatter(x) for x in self.values]
            return fmt_values

    try:
        pf.IntArrayFormatter
        pf.IntArrayFormatter = IntArrayFormatter
    except AttributeError:
        pf._IntArrayFormatter = IntArrayFormatter


def replace_with(string_key: str,
                 replace: str = "()/",
                 replace_with: str = '') -> str:
    """Replace characters in a string with a replacement."""
    for symbol in replace:
        string_key = string_key.replace(symbol, replace_with)
    return string_key


def get_colums_accessor(df: pd.DataFrame,
                        all_lower_case=False) -> omegaconf.OmegaConf:
    """Get an dictionary augmented with attribute access of column name as key
       with white spaces replaced and the original column name as values."""
    cols = {
        replace_with(col.replace(' ', '_').replace('-', '_')): col
        for col in df.columns
    }
    if all_lower_case:
        cols = {k.lower(): v for k, v in cols.items()}
    return omegaconf.OmegaConf.create(cols)


def col_isin_df(cols: typing.Union[list, str], df: pd.DataFrame) -> list:
    """Remove item (column) from passed list if not in DataFrame.
       Warning is issued for missing items.

       cols can be a comma-separated string of column names.
    """
    if isinstance(cols, str):
        cols = cols.split(',')
    ret = list()
    for _var in cols:
        if _var not in df.columns:
            logger.warning(f"Desired variable not found: {_var}", stacklevel=0)
            continue
        ret.append(_var)
    return ret


def value_counts_with_margins(y: pd.Series) -> pd.DataFrame:
    """Value counts of Series with proportion as margins."""
    ret = y.value_counts().to_frame('counts')
    ret.index.name = y.name
    ret['prop'] = y.value_counts(normalize=True)
    return ret


def get_overlapping_columns(df: pd.DataFrame, cols_expected: list) -> list:
    """Get overlapping columns between DataFrame and list of expected columns."""
    ret = df.columns.intersection(cols_expected)
    diff = pd.Index(cols_expected).difference(df.columns)
    if not diff.empty:
        logging.warning(
            f"Some columns are requested, but missing: {diff.to_list()}")
    return ret.to_list()


def combine_value_counts(X: pd.DataFrame, dropna=True) -> pd.DataFrame:
    """Pass a selection of columns to combine it's value counts.

    This performs no checks. Make sure the scale of the variables
    you pass is comparable.

    Parameters
    ----------
    X : pandas.DataFrame
        A DataFrame of several columns with values in a similar range.
    dropna : bool, optional
        Exclude NA values from counting, by default True

    Returns
    -------
    pandas.DataFrame
        DataFrame of combined value counts.
    """
    freq_targets = list()
    for col in X.columns:
        freq_targets.append(X[col].value_counts(dropna=dropna).rename(col))
    freq_targets = pd.concat(freq_targets, axis=1, sort=True)
    return freq_targets



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.4
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Logistic regression model
# Procedure:
#
#
# Example: Alzheimers mass spectrometry-based proteomics dataset
#
# > Predict Alzheimer disease based on proteomics measurements.

# %% tags=["hide-output"]
# Setup colab installation
# You need to restart the runtime after running this cell
# %pip install njab heatmapz openpyxl plotly umap-learn

# %% tags=["hide-input"]
import itertools
import logging
from pathlib import Path
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.express as px
import seaborn
import sklearn
import sklearn.impute
import statsmodels.api as sm
import umap
from heatmap import corrplot
from IPython.display import display
from sklearn.metrics import log_loss, make_scorer

import njab.sklearn
from njab.plotting.metrics import plot_auc, plot_prc
from njab.sklearn import StandardScaler
from njab.sklearn import pca as njab_pca
from njab.sklearn.scoring import (
    ConfusionMatrix,
    get_lr_multiplicative_decomposition,
    get_pred,
    get_score,
    get_target_count_per_bin,
)
from njab.sklearn.types import Splits

logger = logging.getLogger("njab")
logger.setLevel(logging.INFO)

njab.pandas.set_pandas_options()
pd.options.display.min_rows = 10
pd.options.display.max_columns = 20
njab.plotting.set_font_sizes("x-small")
seaborn.set_style("whitegrid")

njab.plotting.set_font_sizes(8)

# %% [markdown]
# ## Set parameters

# %% tags=["parameters"]
CLINIC: str = (
    "https://raw.githubusercontent.com/RasmussenLab/njab/HEAD/docs/tutorial/data/alzheimer/clinic_ml.csv"  # clincial data
)
fname_omics: str = (
    "https://raw.githubusercontent.com/RasmussenLab/njab/HEAD/docs/tutorial/data/alzheimer/proteome.csv"  # omics data
)
TARGET: str = "AD"  # target column in CLINIC dataset (binary)
TARGET_LABEL: Optional[str] = None  # optional: rename target variable
n_features_max: int = 5
freq_cutoff: float = 0.5  # Omics cutoff for sample completeness
VAL_IDS: str = ""  #
VAL_IDS_query: str = ""
weights: bool = True
FOLDER = "alzheimer"
model_name = "all"

# %% [markdown]
# ## Setup
# ### Load data

# %%
clinic = pd.read_csv(CLINIC, index_col=0).convert_dtypes()
cols_clinic = njab.pandas.get_colums_accessor(clinic)
omics = pd.read_csv(fname_omics, index_col=0)

# %% [markdown]
# Data shapes

# %%
omics.shape, clinic.shape

# %% [markdown]
# See how common omics features are and remove feature below choosen frequency cutoff

# %%
ax = omics.notna().sum().sort_values().plot(rot=45)

# %% tags=["hide-input"]

M_before = omics.shape[1]
omics = omics.dropna(thresh=int(len(omics) * freq_cutoff), axis=1)
M_after = omics.shape[1]
msg = (
    f"Removed {M_before-M_after} features with more than {freq_cutoff*100}% missing values."
    f"\nRemaining features: {M_after} (of {M_before})"
)
print(msg)
# keep a map of all proteins in protein group, but only display first protein
# proteins are unique to protein groups
pg_map = {k: k.split(";")[0] for k in omics.columns}
omics = omics.rename(columns=pg_map)
# log2 transform raw intensity data:
omics = np.log2(omics + 1)
omics

# %% [markdown]
# ## Clinical data
# View clinical data

# %%
clinic

# %% [markdown]
# ## Target
# Tabulate target and check for missing values
# %%
njab.pandas.value_counts_with_margins(clinic[TARGET])

# %% tags=["hide-input"]
target_counts = clinic[TARGET].value_counts()

if target_counts.sum() < len(clinic):
    print(
        "Target has missing values."
        f" Can only use {target_counts.sum()} of {len(clinic)} samples."
    )
    mask = clinic[TARGET].notna()
    clinic, omics = clinic.loc[mask], omics.loc[mask]

# %%
if TARGET_LABEL is None:
    TARGET_LABEL = TARGET
y = clinic[TARGET].rename(TARGET_LABEL).astype(int)
clinic_for_ml = clinic.drop(TARGET, axis=1)

# %% [markdown]
# ## Test IDs
# Select some test samples:

# %% tags=["hide-input"]
olink_val, clinic_val = None, None
if not VAL_IDS:
    if VAL_IDS_query:
        logging.warning(f"Querying index using: {VAL_IDS_query}")
        VAL_IDS = clinic.filter(like=VAL_IDS_query, axis=0).index.to_list()
        logging.warning(f"Found {len(VAL_IDS)} Test-IDs")
    else:
        logging.warning("Create train and test split.")
        _, VAL_IDS = sklearn.model_selection.train_test_split(
            clinic.index, test_size=0.2, random_state=123, stratify=clinic[TARGET]
        )
        VAL_IDS = list(VAL_IDS)
elif isinstance(VAL_IDS, str):
    VAL_IDS = VAL_IDS.split(",")
else:
    raise ValueError("Provide IDs in csv format as str: 'ID1,ID2'")
VAL_IDS

# %% [markdown]
# ## Combine clinical and olink data
#

# %% tags=["hide-output"]
# in case you need to subselect
feat_to_consider = clinic_for_ml.columns.to_list()
feat_to_consider += omics.columns.to_list()
feat_to_consider

# %% [markdown]
# View data for training

# %% tags=["hide-input"]
X = clinic_for_ml.join(omics)[feat_to_consider]
X

# %% [markdown]
# ## Data Splits
# Separate train and test split

# %% tags=["hide-input"]
TRAIN_LABEL = "train"
TEST_LABEL = "test"
if VAL_IDS:
    diff = pd.Index(VAL_IDS)
    VAL_IDS = X.index.intersection(VAL_IDS)
    if len(diff) < len(VAL_IDS):
        logging.warning(
            "Some requested validation IDs are not in the data: "
            ",".join(str(x) for x in diff.difference(VAL_IDS))
        )
    X_val = X.loc[VAL_IDS]
    X = X.drop(VAL_IDS)

    use_val_split = True

    y_val = y.loc[VAL_IDS]
    y = y.drop(VAL_IDS)

# %% [markdown]
# ## Output folder

# %% tags=["hide-input"]
FOLDER = Path(FOLDER)
FOLDER.mkdir(exist_ok=True, parents=True)
print(f"Output folder: {FOLDER}")

# %% [markdown]
# ### Outputs
# Save outputs to excel file:

# %% tags=["hide-input"]
# out
files_out = {}
fname = FOLDER / "log_reg.xlsx"
files_out[fname.stem] = fname
writer = pd.ExcelWriter(fname)
print(f"Excel-file for tables: {fname}")

# %% [markdown]
# ## Collect test predictions

# %%
predictions = y_val.to_frame("true")

# %% [markdown]
# ## Fill missing values with training median

# %% tags=["hide-input"]
feat_w_missings = X.isna().sum()
feat_w_missings = feat_w_missings.loc[feat_w_missings > 0]
feat_w_missings

# %% tags=["hide-input"]
row_w_missing = X.isna().sum(axis=1).astype(bool)
col_w_missing = X.isna().sum(axis=0).astype(bool)
X.loc[row_w_missing, col_w_missing]

# %% [markdown]
# Impute using median of training data

# %%
median_imputer = sklearn.impute.SimpleImputer(strategy="median")

X = njab.sklearn.transform_DataFrame(X, median_imputer.fit_transform)
X_val = njab.sklearn.transform_DataFrame(X_val, median_imputer.transform)
assert X.isna().sum().sum() == 0
X.shape, X_val.shape

# %% [markdown]
# ## Principal Components
# on standard normalized training data:

# %% tags=["hide-input"]
scaler = StandardScaler()
X_scaled = scaler.fit_transform(X)

PCs, pca = njab_pca.run_pca(X_scaled, n_components=None)
files_out["var_explained_by_PCs.pdf"] = FOLDER / "var_explained_by_PCs.pdf"
ax = njab_pca.plot_explained_variance(pca)
ax.locator_params(axis="x", integer=True)
njab.plotting.savefig(ax.get_figure(), files_out["var_explained_by_PCs.pdf"])
X_scaled.shape

# %% [markdown]
# Plot first 5 PCs with binary target label annotating each sample::

# %% tags=["hide-input"]
files_out["scatter_first_5PCs.pdf"] = FOLDER / "scatter_first_5PCs.pdf"

fig, axes = plt.subplots(5, 2, figsize=(6, 8), layout="constrained")
PCs.columns = [s.replace("principal component", "PC") for s in PCs.columns]
PCs = PCs.join(y.astype("category"))
up_to = min(PCs.shape[-1], 5)
# https://github.com/matplotlib/matplotlib/issues/25538
# colab: old pandas version and too new matplotlib version (2023-11-6)
for (i, j), ax in zip(itertools.combinations(range(up_to), 2), axes.flatten()):
    PCs.plot.scatter(i, j, c=TARGET_LABEL, cmap="Paired", ax=ax)
_ = PCs.pop(TARGET_LABEL)
njab.plotting.savefig(fig, files_out["scatter_first_5PCs.pdf"])

# %% [markdown]
# ## UMAP
# of training data:

# %% tags=["hide-input"]
reducer = umap.UMAP()
embedding = reducer.fit_transform(X_scaled)

files_out["umap.pdf"] = FOLDER / "umap.pdf"

embedding = pd.DataFrame(
    embedding, index=X_scaled.index, columns=["UMAP 1", "UMAP 2"]
).join(y.astype("category"))
ax = embedding.plot.scatter("UMAP 1", "UMAP 2", c=TARGET_LABEL, cmap="Paired")
njab.plotting.savefig(ax.get_figure(), files_out["umap.pdf"])

# %% [markdown]
# ## Baseline Model - Logistic Regression
# Based on parameters, use weighting:
# %%
if weights:
    weights = "balanced"
    cutoff = 0.5
else:
    cutoff = None
    weights = None

# %% [markdown]
# ## Logistic Regression
# Procedure:
# 1. Select best set of features from entire feature set selected using CV on train split
# 2. Retrain best model configuration using entire train split and evalute on test split
#
# Define splits and models:

# %%
splits = Splits(
    X_train=X_scaled, X_test=scaler.transform(X_val), y_train=y, y_test=y_val
)
model = sklearn.linear_model.LogisticRegression(penalty="l2", class_weight=weights)

# %% tags=["hide-input"]
scoring = [
    "precision",
    "recall",
    "f1",
    "balanced_accuracy",
    "roc_auc",
    "average_precision",
]
scoring = {k: k for k in scoring}
# do not average log loss for AIC and BIC calculations
scoring["log_loss"] = make_scorer(log_loss, greater_is_better=True, normalize=False)
cv_feat = njab.sklearn.find_n_best_features(
    X=splits.X_train,
    y=splits.y_train,
    model=model,
    name=TARGET_LABEL,
    groups=splits.y_train,
    n_features_max=n_features_max,
    scoring=scoring,
    return_train_score=True,
    # fit_params=dict(sample_weight=weights)
)
cv_feat = cv_feat.drop("test_case", axis=1).groupby("n_features").agg(["mean", "std"])
cv_feat

# %% [markdown]
# Add AIC and BIC for model selection

# %% tags=["hide-input"]
# AIC vs BIC on train and test data with bigger is better
IC_criteria = pd.DataFrame()
N_split = {
    "train": round(len(splits.X_train) * 0.8),
    "test": round(len(splits.X_train) * 0.2),
}

for _split in ("train", "test"):

    IC_criteria[(f"{_split}_neg_AIC", "mean")] = -(
        2 * cv_feat.index.to_series() - 2 * cv_feat[(f"{_split}_log_loss", "mean")]
    )
    IC_criteria[(f"{_split}_neg_BIC", "mean")] = -(
        cv_feat.index.to_series() * np.log(N_split[_split])
        - 2 * cv_feat[(f"{_split}_log_loss", "mean")]
    )
IC_criteria.columns = pd.MultiIndex.from_tuples(IC_criteria.columns)
IC_criteria

# %% [markdown]
# All cross-validation metrics:

# %%
cv_feat = cv_feat.join(IC_criteria)
cv_feat = cv_feat.filter(regex="train|test", axis=1).style.highlight_max(
    axis=0, subset=pd.IndexSlice[:, pd.IndexSlice[:, "mean"]]
)
cv_feat

# %% [markdown]
# Save:

# %%
cv_feat.to_excel(writer, sheet_name="CV", float_format="%.3f")
cv_feat = cv_feat.data

# %% [markdown]
# Optimal number of features to use based on cross-validation by metric:

# %% tags=["hide-input"]
mask = cv_feat.columns.levels[0].str[:4] == "test"
scores_cols = cv_feat.columns.levels[0][mask]
n_feat_best = cv_feat.loc[:, pd.IndexSlice[scores_cols, "mean"]].idxmax()
n_feat_best.name = "best"
n_feat_best.to_excel(writer, sheet_name="n_feat_best")
n_feat_best

# %% [markdown]
# Retrain model with best number of features by selected metric::

# %%
results_model = njab.sklearn.run_model(
    model=model,
    splits=splits,
    n_feat_to_select=n_feat_best.loc["test_roc_auc", "mean"],
)
results_model.name = model_name

# %% [markdown]
# ## Receiver Operating Curve of final model

# %% tags=["hide-input"]
ax = plot_auc(
    results_model, label_train=TRAIN_LABEL, label_test=TEST_LABEL, figsize=(4, 2)
)
files_out["ROAUC"] = FOLDER / "plot_roauc.pdf"
njab.plotting.savefig(ax.get_figure(), files_out["ROAUC"])

# %% [markdown]
# ## Precision-Recall Curve for final model

# %% tags=["hide-input"]
ax = plot_prc(
    results_model, label_train=TRAIN_LABEL, label_test=TEST_LABEL, figsize=(4, 2)
)
files_out["PRAUC"] = FOLDER / "plot_prauc.pdf"
njab.plotting.savefig(ax.get_figure(), files_out["PRAUC"])

# %% [markdown]
# ## Coefficients with/out std. errors

# %% tags=["hide-input"]
pd.DataFrame(
    {
        "coef": results_model.model.coef_.flatten(),
        "name": results_model.model.feature_names_in_,
    }
)

# %%
results_model.model.intercept_

# %% [markdown]
# ## Selected Features

# %% tags=["hide-input"]
des_selected_feat = splits.X_train[results_model.selected_features].describe()
des_selected_feat.to_excel(writer, sheet_name="sel_feat", float_format="%.3f")
des_selected_feat

# %% [markdown]
# ### Heatmap of correlations

# %% tags=["hide-input"]
fig = plt.figure(figsize=(6, 6))
files_out["corr_plot_train.pdf"] = FOLDER / "corr_plot_train.pdf"
_ = corrplot(X[results_model.selected_features].join(y).corr(), size_scale=300)
njab.plotting.savefig(fig, files_out["corr_plot_train.pdf"])

# %% [markdown]
# ## Plot training data scores

# %% tags=["hide-input"]
N_BINS = 20
score = get_score(
    clf=results_model.model, X=splits.X_train[results_model.selected_features], pos=1
)
ax = score.hist(bins=N_BINS)
files_out["hist_score_train.pdf"] = FOLDER / "hist_score_train.pdf"
njab.plotting.savefig(ax.get_figure(), files_out["hist_score_train.pdf"])
pred_bins = get_target_count_per_bin(score, y, n_bins=N_BINS)
ax = pred_bins.plot(kind="bar", ylabel="count")
files_out["hist_score_train_target.pdf"] = FOLDER / "hist_score_train_target.pdf"
njab.plotting.savefig(ax.get_figure(), files_out["hist_score_train_target.pdf"])
# pred_bins

# %% [markdown]
# ## Test data scores

# %% tags=["hide-input"]
score_val = get_score(
    clf=results_model.model, X=splits.X_test[results_model.selected_features], pos=1
)
predictions["score"] = score_val
ax = score_val.hist(bins=N_BINS)  # list(x/N_BINS for x in range(0,N_BINS)))
ax.set_ylabel("count")
ax.set_xlim(0, 1)
files_out["hist_score_test.pdf"] = FOLDER / "hist_score_test.pdf"
njab.plotting.savefig(ax.get_figure(), files_out["hist_score_test.pdf"])
pred_bins_val = get_target_count_per_bin(score_val, y_val, n_bins=N_BINS)
ax = pred_bins_val.plot(kind="bar", ylabel="count")
ax.locator_params(axis="y", integer=True)
files_out["hist_score_test_target.pdf"] = FOLDER / "hist_score_test_target.pdf"
njab.plotting.savefig(ax.get_figure(), files_out["hist_score_test_target.pdf"])
# pred_bins_val

# %% [markdown]
# ## Performance evaluations
# Check if the cutoff can be adapted to maximize the F1 score
# between precision and recall:

# %% tags=["hide-input"]
prc = pd.DataFrame(results_model.train.prc, index="precision recall cutoffs".split())
prc

# %% tags=["hide-input"]
prc.loc["f1_score"] = (
    2
    * (prc.loc["precision"] * prc.loc["recall"])
    / (1 / prc.loc["precision"] + 1 / prc.loc["recall"])
)
f1_max = prc[prc.loc["f1_score"].argmax()]
f1_max

# %% [markdown]
# Cutoff set

# %% tags=["hide-input"]
cutoff = float(f1_max.loc["cutoffs"])
cutoff

# %% tags=["hide-input"]
y_pred_val = njab.sklearn.scoring.get_custom_pred(
    clf=results_model.model,
    X=splits.X_test[results_model.selected_features],
    cutoff=cutoff,
)
predictions[model_name] = y_pred_val
predictions["dead"] = y_val
_ = ConfusionMatrix(y_val, y_pred_val).as_dataframe()
_.columns = pd.MultiIndex.from_tuples(
    [(t[0] + f" - {cutoff:.3f}", t[1]) for t in _.columns]
)
_.to_excel(writer, sheet_name="CM_test_cutoff_adapted")
_

# %% tags=["hide-input"]
y_pred_val = get_pred(
    clf=results_model.model, X=splits.X_test[results_model.selected_features]
)
predictions[model_name] = y_pred_val
predictions["dead"] = y_val
_ = ConfusionMatrix(y_val, y_pred_val).as_dataframe()
_.columns = pd.MultiIndex.from_tuples([(t[0] + f" - {0.5}", t[1]) for t in _.columns])
_.to_excel(writer, sheet_name="CM_test_cutoff_0.5")
_

# %% [markdown]
# ## Multiplicative decompositon
# Decompose the model into its components for both splits:

# %% tags=["hide-input"]
components = get_lr_multiplicative_decomposition(
    results=results_model, X=splits.X_train, prob=score, y=y
)
components.to_excel(writer, sheet_name="decomp_multiplicative_train")
components.to_excel(
    writer, sheet_name="decomp_multiplicative_train_view", float_format="%.5f"
)
components.head(10)

# %% tags=["hide-input"]
components_test = get_lr_multiplicative_decomposition(
    results=results_model, X=splits.X_test, prob=score_val, y=y_val
)
components_test.to_excel(writer, sheet_name="decomp_multiplicative_test")
components_test.to_excel(
    writer, sheet_name="decomp_multiplicative_test_view", float_format="%.5f"
)
components_test.head(10)


# %% [markdown]
# ## Plot TP, TN, FP and FN on PCA plot
#
# ### UMAP
# %% tags=["hide-input"]
reducer = umap.UMAP(random_state=42)
# bug: how does UMAP works with only one feature?
# make sure to have two or more features?
M_sel = len(results_model.selected_features)
if M_sel > 1:
    embedding = reducer.fit_transform(X_scaled[results_model.selected_features])

    embedding = pd.DataFrame(
        embedding,
        index=X_scaled.index,
        columns=["UMAP dimension 1", "UMAP dimension 2"],
    ).join(y.astype("category"))
    display(embedding.head(3))
else:
    embedding = None

# %% [markdown]
# Annotate using target variable and predictions:

# %% tags=["hide-input"]
predictions["label"] = predictions.apply(
    lambda x: njab.sklearn.scoring.get_label_binary_classification(
        x["true"], x[model_name]
    ),
    axis=1,
)
mask = predictions[["true", model_name]].sum(axis=1).astype(bool)
predictions.loc[mask].sort_values("score", ascending=False)

# %% tags=["hide-input"]
X_val_scaled = scaler.transform(X_val)
if embedding is not None:
    embedding_val = pd.DataFrame(
        reducer.transform(X_val_scaled[results_model.selected_features]),
        index=X_val_scaled.index,
        columns=["UMAP dimension 1", "UMAP dimension 2"],
    )
    embedding_val.sample(3)

# %% tags=["hide-input"]
pred_train = (
    y.to_frame("true")
    # .join(get_score(clf=results_model.model, X=splits.X_train[results_model.selected_features], pos=1))
    .join(score.rename("score")).join(
        get_pred(
            results_model.model, splits.X_train[results_model.selected_features]
        ).rename(model_name)
    )
)
pred_train["label"] = pred_train.apply(
    lambda x: njab.sklearn.scoring.get_label_binary_classification(
        x["true"], x[model_name]
    ),
    axis=1,
)
pred_train.sample(5)

# %% tags=["hide-cell"]
colors = seaborn.color_palette(n_colors=4)
colors

# %% tags=["hide-input"]
if embedding is not None:
    fig, axes = plt.subplots(1, 2, figsize=(8, 4), sharex=True, sharey=True)
    for _embedding, ax, _title, _model_pred_label in zip(
        [embedding, embedding_val],
        axes,
        [TRAIN_LABEL, TEST_LABEL],
        [pred_train["label"], predictions["label"]],
    ):  # noqa: E129
        ax = seaborn.scatterplot(
            x=_embedding.iloc[:, 0],
            y=_embedding.iloc[:, 1],
            hue=_model_pred_label,
            hue_order=["TN", "TP", "FN", "FP"],
            palette=[colors[0], colors[2], colors[1], colors[3]],
            ax=ax,
        )
        ax.set_title(_title)

    # files_out['pred_pca_labeled'] = FOLDER / 'pred_pca_labeled.pdf'
    # njab.plotting.savefig(fig, files_out['pred_pca_labeled'])

    files_out["umap_sel_feat.pdf"] = FOLDER / "umap_sel_feat.pdf"
    njab.plotting.savefig(ax.get_figure(), files_out["umap_sel_feat.pdf"])

# %% [markdown]
# ### Interactive UMAP plot
# > Not displayed in online documentation

# %% tags=["hide-input"]
if embedding is not None:
    embedding = embedding.join(X[results_model.selected_features])
    embedding_val = embedding_val.join(X_val[results_model.selected_features])
    embedding["label"], embedding_val["label"] = (
        pred_train["label"],
        predictions["label"],
    )
    embedding["group"], embedding_val["group"] = TRAIN_LABEL, TEST_LABEL
    combined_embeddings = pd.concat([embedding, embedding_val])
    combined_embeddings.index.name = "ID"

# %% tags=["hide-input"]
if embedding is not None:
    cols = combined_embeddings.columns

    TEMPLATE = "none"
    defaults = dict(width=800, height=400, template=TEMPLATE)

    fig = px.scatter(
        combined_embeddings.round(3).reset_index(),
        x=cols[0],
        y=cols[1],
        color="label",
        facet_col="group",
        hover_data=["ID"] + results_model.selected_features,
        **defaults,
    )
    fig.for_each_annotation(lambda a: a.update(text=a.text.split("=")[1]))

    fname = FOLDER / "umap_sel_feat.html"
    files_out[fname.name] = fname
    fig.write_html(fname)
    print(fname)
    display(fig)

# %% [markdown]
# ### PCA

# %% tags=["hide-input"]
PCs_train, pca = njab_pca.run_pca(
    X_scaled[results_model.selected_features], n_components=None
)
ax = njab_pca.plot_explained_variance(pca)
ax.locator_params(axis="x", integer=True)

fname = FOLDER / "feat_sel_PCA_var_explained_by_PCs.pdf"
files_out[fname.name] = fname
njab.plotting.savefig(ax.get_figure(), fname)

# %% [markdown]
# Applied to the test split:

# %% tags=["hide-input"]
PCs_val = pca.transform(X_val_scaled[results_model.selected_features])
PCs_val = pd.DataFrame(PCs_val, index=X_val_scaled.index, columns=PCs_train.columns)
PCs_val

# %% tags=["hide-input"]
if M_sel > 1:
    fig, axes = plt.subplots(1, 2, figsize=(6, 3), sharex=True, sharey=True)
    for _embedding, ax, _title, _model_pred_label in zip(
        [PCs_train, PCs_val],
        axes,
        [TRAIN_LABEL, TEST_LABEL],
        [pred_train["label"], predictions["label"]],
    ):  # noqa: E129
        ax = seaborn.scatterplot(
            x=_embedding.iloc[:, 0],
            y=_embedding.iloc[:, 1],
            hue=_model_pred_label,
            hue_order=["TN", "TP", "FN", "FP"],
            palette=[colors[0], colors[2], colors[1], colors[3]],
            ax=ax,
        )
        ax.set_title(_title)

    fname = FOLDER / "pca_sel_feat.pdf"
    files_out[fname.name] = fname
    njab.plotting.savefig(ax.get_figure(), fname)

# %% tags=["hide-input"]
if M_sel > 1:
    max_rows = min(3, len(results_model.selected_features))
    fig, axes = plt.subplots(
        max_rows, 2, figsize=(6, 8), sharex=False, sharey=False, layout="constrained"
    )

    for axes_col, (_embedding, _title, _model_pred_label) in enumerate(
        zip(
            [PCs_train, PCs_val],
            [TRAIN_LABEL, TEST_LABEL],
            [pred_train["label"], predictions["label"]],
        )
    ):
        _row = 0
        axes[_row, axes_col].set_title(_title)
        for i, j in itertools.combinations(range(max_rows), 2):
            ax = seaborn.scatterplot(
                x=_embedding.iloc[:, i],
                y=_embedding.iloc[:, j],
                hue=_model_pred_label,
                hue_order=["TN", "TP", "FN", "FP"],
                palette=[colors[0], colors[2], colors[1], colors[3]],
                ax=axes[_row, axes_col],
            )
            _row += 1

    fname = FOLDER / f"pca_sel_feat_up_to_{max_rows}.pdf"
    files_out[fname.name] = fname
    njab.plotting.savefig(ax.get_figure(), fname)

# %% [markdown]
# ### Features
# - top 3 scaled n_features_max (scatter)
# - or unscalled single features (swarmplot)

# %% tags=["hide-input"]
if M_sel > 1:
    max_rows = min(3, len(results_model.selected_features))
    fig, axes = plt.subplots(
        max_rows, 2, figsize=(6, 8), sharex=False, sharey=False, layout="constrained"
    )

    for axes_col, (_embedding, _title, _model_pred_label) in enumerate(
        zip(
            [
                X_scaled[results_model.selected_features],
                X_val_scaled[results_model.selected_features],
            ],
            [TRAIN_LABEL, TEST_LABEL],
            [pred_train["label"], predictions["label"]],
        )
    ):
        _row = 0
        axes[_row, axes_col].set_title(_title)
        for i, j in itertools.combinations(range(max_rows), 2):
            ax = seaborn.scatterplot(
                x=_embedding.iloc[:, i],
                y=_embedding.iloc[:, j],
                hue=_model_pred_label,
                hue_order=["TN", "TP", "FN", "FP"],
                palette=[colors[0], colors[2], colors[1], colors[3]],
                ax=axes[_row, axes_col],
            )
            _row += 1

    fname = FOLDER / f"sel_feat_up_to_{max_rows}.pdf"
    files_out[fname.name] = fname
    njab.plotting.savefig(ax.get_figure(), fname)
else:
    fig, axes = plt.subplots(1, 1, figsize=(6, 2), layout="constrained")
    single_feature = results_model.selected_features[0]
    data = pd.concat(
        [
            X[single_feature]
            .to_frame()
            .join(pred_train["label"])
            .assign(group=TRAIN_LABEL),
            X_val[single_feature]
            .to_frame()
            .join(predictions["label"])
            .assign(group=TEST_LABEL),
        ]
    )
    ax = seaborn.swarmplot(data=data, x="group", y=single_feature, hue="label", ax=axes)
    fname = FOLDER / f"sel_feat_{single_feature}.pdf"
    files_out[fname.name] = fname
    njab.plotting.savefig(ax.get_figure(), fname)

# %% [markdown]
# ## Savee annotation of errors for manuel analysis
#
# Saved to excel table.

# %%
X[results_model.selected_features].join(pred_train).to_excel(
    writer, sheet_name="pred_train_annotated", float_format="%.3f"
)
X_val[results_model.selected_features].join(predictions).to_excel(
    writer, sheet_name="pred_test_annotated", float_format="%.3f"
)

# %% [markdown]
# ## Outputs

# %%
writer.close()
files_out



import pandas as pd
from sklearn import preprocessing


class StandardScaler(preprocessing.StandardScaler):
    """Standardscaler which keeps column names and indices of pandas DataFrames."""

    def transform(self, X, copy=None):
        res = super().transform(X, copy)
        if isinstance(X, pd.DataFrame):
            return pd.DataFrame(res, columns=X.columns, index=X.index)
        return res

    def inverse_transform(self, X, copy=None):
        res = super().inverse_transform(X, copy)
        if isinstance(X, pd.DataFrame):
            return pd.DataFrame(res, columns=X.columns, index=X.index)
        return res



import numpy as np
import numpy.testing as npt
import pandas as pd
import sklearn.preprocessing as preprocessing

from njab.sklearn.preprocessing import StandardScaler


def test_StandardScaler():
    X = pd.DataFrame(np.array([[2, None], [3, 2], [4, 6]]))
    npt.assert_almost_equal(preprocessing.StandardScaler().fit(X).transform(X),
                            StandardScaler().fit(X).transform(X).to_numpy())



# %%
# pip install SurvSet

# %%
from pathlib import Path

from SurvSet.data import SurvLoader

loader = SurvLoader()
# List of available datasets and meta-info
loader.df_ds.head()

# %%
df, ref = loader.load_dataset(ds_name='prostate').values()

# %%
fname = "prostate.csv"

df.to_csv(fname, index=False)



import matplotlib
import matplotlib.pyplot as plt
import pandas as pd

from njab.sklearn.types import ResultsSplit, Results

LIMITS = (-0.05, 1.05)


def plot_split_auc(result: ResultsSplit, name: str,
                   ax: matplotlib.axes.Axes) -> matplotlib.axes.Axes:
    """Add receiver operation curve to ax of a split of the data."""
    col_name = f"{name} (auc: {result.auc:.3f})"
    roc = pd.DataFrame(result.roc, index='fpr tpr cutoffs'.split()).rename(
        {'tpr': col_name})
    ax = roc.T.plot('fpr',
                    col_name,
                    xlabel='false positive rate',
                    ylabel='true positive rate',
                    style='.-',
                    ylim=LIMITS,
                    xlim=LIMITS,
                    ax=ax)
    return ax


# ! should be roc
def plot_auc(results: Results,
             ax: matplotlib.axes.Axes = None,
             label_train='train',
             label_test='test',
             **kwargs) -> matplotlib.axes.Axes:
    """Plot ROC curve for train and test data."""
    if ax is None:
        fig, ax = plt.subplots(1, 1, **kwargs)
    ax = plot_split_auc(results.train, f"{label_train}", ax)
    ax = plot_split_auc(results.test, f"{label_test}", ax)
    return ax


def plot_split_prc(result: ResultsSplit, name: str,
                   ax: matplotlib.axes.Axes) -> matplotlib.axes.Axes:
    """Add precision recall curve to ax of a split of the data."""
    col_name = f"{name} (aps: {result.aps:.3f})"
    roc = pd.DataFrame(result.prc,
                       index='precision recall cutoffs'.split()).rename(
                           {'precision': col_name})
    ax = roc.T.plot('recall',
                    col_name,
                    xlabel='true positive rate',
                    ylabel='precision',
                    style='.-',
                    ylim=LIMITS,
                    xlim=LIMITS,
                    ax=ax)
    return ax


def plot_prc(results: ResultsSplit,
             ax: matplotlib.axes.Axes = None,
             label_train='train',
             label_test='test',
             **kwargs):
    """Plot precision recall curve for train and test data."""
    if ax is None:
        fig, ax = plt.subplots(1, 1, **kwargs)
    ax = plot_split_prc(results.train, f"{label_train}", ax)
    ax = plot_split_prc(results.test, f"{label_test}", ax)
    return ax



"""Scikit-learn related functionality.

Finds the a good set of features using minimum redundancy maximum relevance (MRMR)
for a logistic regression model a binary target variable.
"""
from typing import Optional

import pandas as pd
import sklearn
import sklearn.model_selection

from njab.sklearn import scoring
from njab.sklearn.pca import run_pca
from njab.sklearn.preprocessing import StandardScaler
from njab.sklearn.types import (AucRocCurve, PrecisionRecallCurve, Results,
                                ResultsSplit, Splits)

__all__ = [
    'run_model',
    'get_results_split',
    'find_n_best_features',
    'run_pca',
    'scoring',
    'StandardScaler',
]

RANDOM_STATE = 42

default_log_reg = sklearn.linear_model.LogisticRegression(
    random_state=RANDOM_STATE
    #   , solver='liblinear'
)


def run_model(
    splits: Splits,
    model: sklearn.base.BaseEstimator = default_log_reg,
    fit_params=None,
    n_feat_to_select=9,
) -> Results:
    """Fit a model on the training split and calculate
       performance metrics on both train and test split.
    """
    from mrmr import mrmr_classif
    selected_features = mrmr_classif(X=splits.X_train,
                                     y=splits.y_train,
                                     K=n_feat_to_select)

    if fit_params is None:
        fit_params = {}

    model = model.fit(splits.X_train[selected_features], splits.y_train,
                      **fit_params)

    pred_score_test = model.predict_proba(splits.X_test[selected_features])[:,
                                                                            1]
    results_test = get_results_split(y_true=splits.y_test,
                                     y_score=pred_score_test)

    pred_score_train = model.predict_proba(
        splits.X_train[selected_features])[:, 1]
    results_train = get_results_split(y_true=splits.y_train,
                                      y_score=pred_score_train)

    ret = Results(model=model,
                  selected_features=selected_features,
                  train=results_train,
                  test=results_test)
    return ret


def get_results_split(y_true, y_score):
    """Calculate metrics for a single set of samples."""
    ret = ResultsSplit(
        auc=sklearn.metrics.roc_auc_score(y_true=y_true, y_score=y_score))

    ret.roc = AucRocCurve(
        *sklearn.metrics.roc_curve(y_true=y_true, y_score=y_score))
    ret.prc = PrecisionRecallCurve(*sklearn.metrics.precision_recall_curve(
        y_true=y_true, probas_pred=y_score))

    ret.aps = sklearn.metrics.average_precision_score(y_true=y_true,
                                                      y_score=y_score)
    return ret


default_log_reg = sklearn.linear_model.LogisticRegression(
    random_state=RANDOM_STATE, solver='liblinear')


def find_n_best_features(
        X: pd.DataFrame,
        y: pd.Series,
        name: str,
        model: sklearn.base.BaseEstimator = default_log_reg,
        groups=None,  # ? Optional[array-like]
        n_features_max: int = 15,
        random_state: int = RANDOM_STATE,
        scoring: Optional[tuple] = ('precision', 'recall', 'f1',
                                    'balanced_accuracy', 'roc_auc',
                                    'average_precision'),
        return_train_score: bool = False,
        fit_params: Optional[dict] = None):
    """Create a summary of model performance on 10 times 5-fold cross-validation."""
    from mrmr import mrmr_classif
    summary = []
    cv = sklearn.model_selection.RepeatedStratifiedKFold(
        n_splits=5, n_repeats=10, random_state=random_state)
    in_both = y.index.intersection(X.index)
    # could have a warning in case
    _X = X.loc[in_both]
    _y = y.loc[in_both]
    n_features_max = min(n_features_max, X.shape[-1])
    for n_features in range(1, n_features_max + 1):
        selected_features = mrmr_classif(_X, _y, K=n_features)
        _X_mrmr = _X[selected_features]
        scores = sklearn.model_selection.cross_validate(
            estimator=model,
            X=_X_mrmr,
            y=_y,
            groups=groups,
            scoring=scoring,
            cv=cv,
            return_train_score=return_train_score,
            params=fit_params,
            error_score='raise')
        scores['n_features'] = n_features
        scores['test_case'] = name
        scores['n_observations'] = _X.shape[0]
        results = pd.DataFrame(scores)
        summary.append(results)
    summary_n_features = pd.concat(summary)
    return summary_n_features


def transform_DataFrame(X: pd.DataFrame, fct: callable) -> pd.DataFrame:
    """Set index and columns of a DataFrame after applying a callable
    which might only return a numpy array.

    Parameters
    ----------
    X : pd.DataFrame
        Original DataFrame to be transformed
    fct : callable
        Callable to be applied to every element in the DataFrame.

    Returns
    -------
    pd.DataFrame
        Transformed DataFrame
    """
    ret = fct(X)
    ret = pd.DataFrame(ret, index=X.index, columns=X.columns)
    return ret



"""Matplotlib functionality for custom plots."""

import logging
import pathlib
from typing import Iterable

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from njab.plotting.lifelines import plot_lifelines

__all__ = ['plot_lifelines']

plt.rcParams['figure.figsize'] = [4.0, 3.0]
plt.rcParams['pdf.fonttype'] = 42
plt.rcParams['ps.fonttype'] = 42

plt.rcParams['figure.dpi'] = 147

figsize_a4 = (8.3, 11.7)

logger = logging.getLogger(__name__)


def savefig(fig: matplotlib.figure.Figure,
            name: str,
            folder: str = '.',
            dpi: int = 300,
            pdf: bool = True,
            tight_layout: bool = True) -> None:
    """Save matplotlib Figure (having method `savefig`) as pdf and png."""
    folder = pathlib.Path(folder)
    fname = folder / name
    folder = fname.parent  # in case name specifies folders
    folder.mkdir(exist_ok=True, parents=True)
    if tight_layout and not fig.get_constrained_layout():
        fig.tight_layout()
    fname = fname.with_suffix('.png')
    fig.savefig(fname, dpi=dpi, bbox_inches='tight')
    logger.info("Saved Figures to %s", fname)
    if pdf:
        fname = fname.with_suffix('.pdf')
        fig.savefig(fname, bbox_inches='tight')
        logger.info("Saved Figures to %s", fname)


def select_xticks(ax: matplotlib.axes.Axes, max_ticks: int = 50) -> list:
    """Limit the number of xticks displayed.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Axes object to manipulate
    max_ticks : int, optional
        maximum number of set ticks on x-axis, by default 50

    Returns
    -------
    list
        list of current ticks for x-axis. Either new
        or old (depending if something was changed).
    """
    x_ticks = ax.get_xticks()
    offset = len(x_ticks) // max_ticks
    if offset > 1:  # if larger than 1
        return ax.set_xticks(x_ticks[::offset])
    return x_ticks


def select_dates(date_series: pd.Series, max_ticks=30) -> np.array:
    """Get unique dates (single days) for selection in pd.plot.line
    with xticks argument.

    Parameters
    ----------
    date_series : pd.Series
        datetime series to use (values, not index)
    max_ticks : int, optional
        maximum number of unique ticks to select, by default 30

    Returns
    -------
    np.array
        array of selected dates
    """
    xticks = date_series.dt.date.unique()
    offset = len(xticks) // max_ticks
    if offset > 1:
        return xticks[::offset]
    else:
        xticks


def make_large_descriptors(size='xx-large'):
    """Helper function to have very large titles, labes and tick texts for
    matplotlib plots per default.

    size: str
        fontsize or allowed category. Change default if necessary, default 'xx-large'
    """
    plt.rcParams.update({
        k: size
        for k in [
            'xtick.labelsize', 'ytick.labelsize', 'axes.titlesize',
            'axes.labelsize', 'legend.fontsize', 'legend.title_fontsize'
        ]
    })


set_font_sizes = make_large_descriptors


def add_prop_as_second_yaxis(
        ax: matplotlib.axes.Axes,
        n_samples: int,
        format_str: str = '{x:,.3f}') -> matplotlib.axes.Axes:
    """Add proportion as second axis. Try to align cleverly

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Axes for which you want to add a second y-axis
    n_samples : int
        Number of total samples (to normalize against)

    Returns
    -------
    matplotlib.axes.Axes
        Second layover twin Axes with right-hand side y-axis
    """
    ax2 = ax.twinx()
    n_min, n_max = np.round(ax.get_ybound())
    logger.info(f"{n_min = }, {n_max = }")
    lower_prop = n_min / n_samples + (ax.get_ybound()[0] - n_min) / n_samples
    upper_prop = n_max / n_samples + (ax.get_ybound()[1] - n_max) / n_samples
    logger.info(f'{lower_prop = }, {upper_prop = }')
    ax2.set_ybound(lower_prop, upper_prop)
    # _ = ax2.set_yticks(np.linspace(n_min/n_samples,
    #                    n_max /n_samples, len(ax.get_yticks())-2))
    _ = ax2.set_yticks(ax.get_yticks()[1:-1] / n_samples)
    ax2.yaxis.set_major_formatter(
        matplotlib.ticker.StrMethodFormatter(format_str))
    return ax2


def add_height_to_barplot(ax: matplotlib.axes.Axes,
                          size: int = 15) -> matplotlib.axes.Axes:
    """Add height of bar to each bar in a barplot."""
    for bar in ax.patches:
        ax.annotate(text=format(bar.get_height(), '.2f'),
                    xy=(bar.get_x() + bar.get_width() / 2, bar.get_height()),
                    xytext=(0, 7),
                    ha='center',
                    va='center',
                    size=size,
                    textcoords='offset points')
    return ax


def add_text_to_barplot(ax: matplotlib.axes.Axes,
                        text: Iterable[str],
                        size=15) -> matplotlib.axes.Axes:
    """Add custom text from Iterable to each bar in a barplot."""
    for bar, text_bar in zip(ax.patches, text):
        logger.debug(f"{bar = }, f{text = }, {bar.get_height() = }")
        ax.annotate(text=text_bar,
                    xy=(bar.get_x() + bar.get_width() / 2, bar.get_height()),
                    xytext=(0, -5),
                    rotation=90,
                    ha='center',
                    va='top',
                    size=size,
                    textcoords='offset points')
    return ax


def format_large_numbers(ax: matplotlib.axes.Axes,
                         format_str: str = '{x:,.0f}') -> matplotlib.axes.Axes:
    """Format large integer numbers to be read more easily.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Axes which labels should be manipulated.
    format_str : str, optional
        Default float format string, by default '{x:,.0f}'

    Returns
    -------
    matplotlib.axes.Axes
        Return reference to modified input Axes object.
    """
    ax.xaxis.set_major_formatter(
        matplotlib.ticker.StrMethodFormatter(format_str))
    ax.yaxis.set_major_formatter(
        matplotlib.ticker.StrMethodFormatter(format_str))
    return ax



# %%
from pathlib import Path
import pandas as pd

# %%
fname_curated_data = 'Alzheimer.xlsx'

FOLDER = Path('alzheimer').absolute()
FOLDER.mkdir(exist_ok=True, parents=True)
files_out = dict()

# %%
data = pd.read_excel(fname_curated_data)
data.index = [f'Sample_{i:03d}' for i in data.index]
data.index.name = 'Sample ID'
data.head()

# %%
meta = data.filter(like='_', axis=1)
meta.sample(5)

# %% [markdown]
# # Meta data dump
# - N=197 samples with complete data for
#   - age
#   - gender
#   - primary biochemical Alzheimer disease classification

# %%
cols = [
    '_age at CSF collection',
    '_gender',
    '_primary biochemical AD classification'
]
meta[cols].dropna().describe(include='all')


# %%
y_obj = meta['_primary biochemical AD classification'].rename('AD')
y = pd.get_dummies(y_obj)["biochemical AD"].rename('AD').to_frame()
y

# %% [markdown]
# Encode some clincial variables as binary

# %% tags=["hide-input"]
# ! preprocessing should move to setup script
dummies_collection_site = pd.get_dummies(
    meta['_collection site'])  # reference is first column (Berlin)
dummies_collection_site.describe()

# %% tags=["hide-input"]
clinic_for_ml = dummies_collection_site.iloc[:, 1:]
clinic_for_ml = (clinic_for_ml.join(
    pd.get_dummies(meta['_gender']).rename(columns={
        'f': 'female',
        'm': 'male'
    }).iloc[:, 1:])).assign(age=meta['_age at CSF collection']).join(
        y
)
# clinic_for_ml = clinic[["EGFR", "M"]]
clinic_for_ml

# %%
fname = FOLDER / 'clinic_ml.csv'
files_out[fname.stem] = fname
clinic_for_ml.to_csv(fname)

# %%
fname = FOLDER / 'meta.csv'
files_out[fname.stem] = fname
meta.to_csv(fname)

# %% [markdown]
# # Proteome dump
# %%
data = data.iloc[:, :-11]
data.sample(5)

# %%
ax = data.notna().sum(axis=0).plot.box()

# %%
fname = FOLDER / 'proteome.csv'
files_out[fname.stem] = fname
data.to_csv(fname)

# %% [markdown]
# Protein Group - Gene Mapping

# %%

protein_groups = data.columns.to_list()
proteins_unique = set()
for pg in protein_groups:
    proteins_unique |= set(pg.split(';'))
# proteins_unique

# %%
files_out

# %%



"""not just another biomarker (njab) - a package for biomarker discovery

Provides an opinionated workflow for biomarker discovery, using
mimimum redundancy maximum relevance (MRMR) feature selection and
logistic regression as a simple and explainable model.
"""
from importlib.metadata import version

from njab import io, pandas, plotting, sklearn, stats

__version__ = version('njab')

__all__ = ['stats', 'sklearn', 'plotting', 'pandas', 'io']



from typing import Optional

import matplotlib
import pandas as pd
import sklearn.decomposition


def run_pca(
        df_wide: pd.DataFrame,
        n_components: int = 2
) -> tuple[pd.DataFrame, sklearn.decomposition.PCA]:
    """Run PCA on DataFrame and return result.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame in wide format to fit features on.
    n_components : int, optional
        Number of Principal Components to fit, by default 2

    Returns
    -------
    Tuple[pd.DataFrame, PCA]
        principal compoments of DataFrame with same indices as in original DataFrame,
        and fitted PCA model of sklearn
    """
    n_comp_max = None
    if n_components is not None:
        n_comp_max = min(df_wide.shape)
        n_comp_max = min(n_comp_max, n_components)
    pca = sklearn.decomposition.PCA(n_components=n_comp_max)
    PCs = pca.fit_transform(df_wide)
    cols = [
        f'principal component {i+1} ({var_explained*100:.2f} %)'
        for i, var_explained in enumerate(pca.explained_variance_ratio_)
    ]
    PCs = pd.DataFrame(PCs, index=df_wide.index, columns=cols)
    return PCs, pca


def plot_explained_variance(
        pca: sklearn.decomposition.PCA,
        ax: Optional[matplotlib.axes.Axes] = None) -> matplotlib.axes.Axes:
    """Plot explained variance of PCA from scikit-learn."""
    exp_var = pd.Series(
        pca.explained_variance_ratio_).to_frame('explained variance')
    exp_var.index += 1  # start at 1
    exp_var["explained variance (cummulated)"] = exp_var[
        'explained variance'].cumsum()
    exp_var.index.name = 'PC'
    ax = exp_var.plot(ax=ax)
    return ax



"""Statistical functionalities for njab.

Analysis of covariance, binomial testing for binary variables,
and differential analysis of continuous variables between groups
using t-tests.
"""
from njab.stats import ancova
from njab.stats.groups_comparision import binomtest, diff_analysis

__all__ = ['ancova', 'diff_analysis', 'binomtest']



"""Bionomial test and t-test for groups comparision."""
import logging
import pandas as pd
import pingouin as pg

from scipy.stats import binomtest as scipy_binomtest

logger = logging.getLogger(__name__)


def means_between_groups(
    df: pd.DataFrame,
    boolean_array: pd.Series,
    event_names: tuple[str, str] = ('1', '0')
) -> pd.DataFrame:
    """Mean comparison between groups"""
    sub = df.loc[boolean_array].describe().iloc[:3]
    sub['event'] = event_names[0]
    sub = sub.set_index('event', append=True).swaplevel()
    ret = sub
    sub = df.loc[~boolean_array].describe().iloc[:3]
    sub['event'] = event_names[1]
    sub = sub.set_index('event', append=True).swaplevel()
    ret = pd.concat([ret, sub])
    ret.columns.name = 'variable'
    ret.index.names = ('event', 'stats')
    return ret.T


def calc_stats(df: pd.DataFrame, boolean_array: pd.Series,
               vars: list[str]) -> pd.DataFrame:
    """Calculate t-test for each variable in `vars` between two groups defined
    by boolean array."""
    ret = []
    for var in vars:
        _ = pg.ttest(df.loc[boolean_array, var], df.loc[~boolean_array, var])
        ret.append(_)
    ret = pd.concat(ret)
    ret = ret.set_index(vars)
    ret.columns.name = 'ttest'
    ret.columns = pd.MultiIndex.from_product([['ttest'], ret.columns],
                                             names=('test', 'var'))
    return ret


def diff_analysis(
    df: pd.DataFrame,
    boolean_array: pd.Series,
    event_names: tuple[str, str] = ('1', '0'),
    ttest_vars=("alternative", "p-val", "cohen-d")
) -> pd.DataFrame:
    """Differential analysis procedure between two groups. Calculaes
    mean per group and t-test for each variable in `vars` between two groups."""
    ret = means_between_groups(df,
                               boolean_array=boolean_array,
                               event_names=event_names)
    ttests = calc_stats(df, boolean_array=boolean_array, vars=ret.index)
    ret = ret.join(ttests.loc[:, pd.IndexSlice[:, ttest_vars]])
    return ret


def binomtest(
    var: pd.Series,
    boolean_array: pd.Series,
    alternative='two-sided',
    event_names: tuple[str, str] = ('event', 'no-event')
) -> pd.DataFrame:
    """Binomial test for categorical variable between two groups defined by a
    boolean array."""
    entry = {}
    entry['variable'] = var.name

    if var.dtype != 'category':
        logger.warn(
            f"Passed on categorical data (which was expected): {var.name}")
        var = var.astype('category')

    assert len(
        var.cat.categories
    ) == 2, f"No binary variable, found {len(var.cat.categories)} categories: {list(var.cat.categories)}"

    p_1 = var.loc[boolean_array].dropna().cat.codes.mean()

    p_0 = var.loc[~boolean_array].dropna().cat.codes.mean()
    logger.debug(f"p cat==0: {p_0}, p cat==1: {p_1}")

    cat_at_pos_one = var.cat.categories[1]
    logger.debug('Category with code 1', cat_at_pos_one)

    counts = var.loc[boolean_array].value_counts()
    k, n = counts.loc[cat_at_pos_one], counts.sum()

    entry[event_names[0]] = dict(count=n, p=p_1)
    entry[event_names[1]] = dict(
        count=var.loc[~boolean_array].value_counts().sum(), p=p_0)

    test_res = scipy_binomtest(k, n, p_0, alternative=alternative)
    test_res = pd.Series(test_res.__dict__).to_frame('binomial test').unstack()
    test_res.name = entry['variable']
    test_res = test_res.to_frame().T

    entry = pd.DataFrame(entry).set_index('variable', append=True).unstack(0)
    entry = entry.join(test_res)
    return entry



import pandas as pd

import njab


def test_thousands_display():
    njab.pandas.set_pandas_options()
    s = pd.Series([1_000_000])
    assert str(s)[4:13] == '1,000,000'


def test_combine_value_counts():
    df = pd.DataFrame({'a': [1, 2, 2, 2, 3, 3, 3], 'b': [1, 1, 1, 2, 2, 3, 3]})
    exp = {'a': {1: 1, 2: 3, 3: 3}, 'b': {1: 3, 2: 2, 3: 2}}
    act = njab.pandas.combine_value_counts(df).to_dict()
    assert act == exp



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.4
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Explorative Analysis
# Uses the `prostate` time series dataset provided by the `SurvSet` package.
# - [SurvSet package](https://github.com/ErikinBC/SurvSet/tree/main)
# First install the dependencies:

# %% tags=["hide-output"]
# %pip install 'njab[all]' openpyxl

import logging

# %% tags=["hide-cell"]
from functools import partial
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn
import sklearn
from IPython.display import display
from lifelines.plotting import add_at_risk_counts

import njab
import njab.plotting
from njab.plotting.km import compare_km_curves, log_rank_test

njab.pandas.set_pandas_options()
pd.options.display.min_rows = 10
njab.plotting.set_font_sizes("x-small")
seaborn.set_style("whitegrid")
plt.rcParams["figure.figsize"] = [4.0, 4.0]

# %% [markdown]
# ### Set parameters

# %% tags=["parameters"]
TARGET = "event"
TIME_KM = "time"
FOLDER = "prostate"
CLINIC = "https://raw.githubusercontent.com/RasmussenLab/njab/main/docs/tutorial/data/prostate/prostate.csv"
val_ids: str = ""  # List of comma separated values or filepath
#
# list or string of csv, eg. "var1,var2"
clinic_cont = ["age"]
# list or string of csv, eg. "var1,var2"
clinic_binary = ["male", "AD"]
# List of comma separated values or filepath
da_covar = "num_age,num_wt"

# %% tags=["hide-input"]
print(f"Time To Event: {TIME_KM}")
print(f"event (target) variable: {TARGET}")

# %% tags=["hide-cell"]
FOLDER = Path(FOLDER)
FOLDER.mkdir(exist_ok=True, parents=True)
FOLDER
# %% [markdown]
# Inspect the data:

# %% tags=["hide-input"]
clinic = pd.read_csv(CLINIC, index_col=0).dropna(how="any")
clinic.columns.name = "feat_name"  # ! check needs to be implemented
cols_clinic = njab.pandas.get_colums_accessor(clinic)
clinic = clinic.astype(
    {
        var: "int"
        for var in [
            "event",
            "time",
            "num_age",
            "num_wt",
            "num_sbp",
            "num_dbp",
            "num_sz",
            "num_sg",
            "num_sdate",
            "fac_stage",
        ]
    }
)
clinic

# %% [markdown]
# Descriptive statistics of non-numeric variables:

# %%
clinic.describe(include="object")

# %% [markdown]
# Set the binary variables and convert them to categories:
# %%
vars_binary = ["fac_hx", "fac_bm"]
clinic[vars_binary] = clinic[vars_binary].astype("category")

# %% [markdown]
# Covariates to adjust for:

# %% tags=["hide-input"]
check_isin_clinic = partial(njab.pandas.col_isin_df, df=clinic)
covar = check_isin_clinic(da_covar)
covar

# %% [markdown]
# Set continous variables

# %%
vars_cont = [
    "num_age",
    "num_wt",
    "num_sbp",
    "num_dbp",
    "num_hg",
    "num_sz",
    "num_sg",
    "num_ap",
    "num_sdate",
    "fac_stage",
]

# %% [markdown]
# ### Collect outputs
# in an excel file:

# %% tags=["hide-input"]
fname = FOLDER / "1_differential_analysis.xlsx"
files_out = {fname.name: fname}
writer = pd.ExcelWriter(fname)
print(f"Output will be written to: {fname}")

# %% [markdown]
# ## Differences between groups defined by target
# Perform an uncontrolled t-test for continous
# and binomal test for binary variables.
# Besides the test results, summary statistics of both groups are provided.
#
# First, set the binary target as a boolean mask:

# %%
happend = clinic[TARGET].astype(bool)

# %% [markdown]
# ### Univariate t-test for continous variables

# %% tags=["hide-input"]
ana_differential = njab.stats.groups_comparision.diff_analysis(
    clinic[vars_cont],
    happend,
    event_names=(TARGET, "no event"),
)
ana_differential = ana_differential.sort_values(("ttest", "p-val"))
ana_differential.to_excel(writer, sheet_name="clinic continous", float_format="%.4f")
ana_differential

# %% [markdown]
# ### Binomal test for binary variables:

# %% tags=["hide-input"]
diff_binomial = []
for var in vars_binary:
    if len(clinic[var].cat.categories) == 2:
        diff_binomial.append(
            njab.stats.groups_comparision.binomtest(
                clinic[var], happend, event_names=(TARGET, "no-event")
            )
        )
    else:
        logging.warning(
            f"Non-binary variable: {var} with {len(clinic[var].cat.categories)} categories"
        )

diff_binomial = pd.concat(diff_binomial).sort_values(("binomial test", "pvalue"))
diff_binomial.to_excel(writer, sheet_name="clinic binary", float_format="%.4f")
with pd.option_context("display.max_rows", len(diff_binomial)):
    display(diff_binomial)

# %% [markdown]
# ## Analaysis of covariance (ANCOVA)
# Next, we select continous variables controlling for covariates.
#
# First the summary statistics for the target and covariates:

# %% tags=["hide-input"]
clinic_ancova = [TARGET, *covar]
clinic_ancova = clinic[clinic_ancova].copy()
clinic_ancova.describe(include="all")

# %% [markdown]
# Discard all rows with a missing features (if present):

# %% tags=["hide-input"]
clinic_ancova = clinic_ancova.dropna()
categorical_columns = clinic_ancova.columns[clinic_ancova.dtypes == "category"]
print("Available covariates: " ", ".join(categorical_columns.to_list()))
for categorical_column in categorical_columns:
    # only works if no NA and only binary variables!
    clinic_ancova[categorical_column] = clinic_ancova[categorical_column].cat.codes

desc_ancova = clinic_ancova.describe()
desc_ancova.to_excel(writer, sheet_name="covars", float_format="%.4f")
desc_ancova

# %% [markdown]
# Remove non-varying variables (if present):

# %% tags=["hide-input"]
if (desc_ancova.loc["std"] < 0.001).sum():
    non_varying = desc_ancova.loc["std"] < 0.001
    non_varying = non_varying[non_varying].index
    print("Non varying columns: ", ", ".join(non_varying))
    clinic_ancova = clinic_ancova.drop(non_varying, axis=1)
    for col in non_varying:
        covar.remove(col)

# %% [markdown]
# Run ANCOVA:

# %% tags=["hide-input"]
ancova = njab.stats.ancova.AncovaOnlyTarget(
    df_proteomics=clinic[vars_cont].drop(covar, axis=1),
    df_clinic=clinic_ancova,
    target=TARGET,
    covar=covar,
    value_name="",
)
ancova = ancova.ancova().sort_values("p-unc")
ancova = ancova.loc[:, "p-unc":]
ancova.columns = pd.MultiIndex.from_product(
    [["ancova"], ancova.columns], names=("test", "var")
)
ancova.to_excel(writer, sheet_name="olink controlled", float_format="%.4f")
ancova.head(20)

# %% tags=["hide-input"]
writer.close()

# %% [markdown]
# ## Kaplan-Meier (KM) plot for top markers
# Cutoff is defined using a univariate logistic regression
#
#
# $$ \ln \frac{p}{1-p} = \beta_0 + \beta_1 \cdot x $$
# the default cutoff `p=0.5` corresponds to a feature value of:
#
# $$ x = - \frac{\beta_0}{\beta_1} $$
#
# Optional: The cutoff could be adapted to the prevalence of the target.
#
# List of markers with significant difference between groups as defined by ANCOVA:

# %% tags=["hide-input"]
rejected = ancova.query("`('ancova', 'rejected')` == True")
rejected

# %% [markdown]
# Settings for plots
# %%
class_weight = "balanced"
y_km = clinic[TARGET]
time_km = clinic[TIME_KM]
compare_km_curves = partial(
    compare_km_curves,
    time=time_km,
    y=y_km,
    xlim=(0, 80),
    xlabel="time passed",
    ylabel=f"rate {y_km.name}",
)
log_rank_test = partial(
    log_rank_test,
    time=time_km,
    y=y_km,
)
TOP_N = 2  # None = all

# %% [markdown]
# Plot KM curves for `TOP_N` (here two) makers with log-rank test p-value:

# %% tags=["hide-input"]
for marker, _ in rejected.index[:TOP_N]:  # first case done above currently
    fig, ax = plt.subplots()
    class_weight = "balanced"
    # class_weight=None
    model = sklearn.linear_model.LogisticRegression(class_weight=class_weight)
    model = model.fit(X=clinic[marker].to_frame(), y=happend)
    print(
        f"Intercept {float(model.intercept_.squeeze()):5.3f}, coef.: {float(model.coef_.squeeze()):5.3f}"
    )
    # ! could be adapted based on proportion of target (for imbalanced data):
    # offset = np.log(p/(1-p))
    offset = np.log(0.5 / (1 - 0.5))  # ! standard cutoff of probability of 0.5
    cutoff = offset - float(model.intercept_.squeeze()) / float(model.coef_.squeeze())
    direction = ">" if model.coef_ > 0 else "<"
    print(f"Custom cutoff defined by Logistic regressor for {marker:>10}: {cutoff:.3f}")
    pred = njab.sklearn.scoring.get_pred(model, clinic[marker].to_frame())
    ax, kmf_0, kmf_1 = compare_km_curves(pred=pred)
    res = log_rank_test(mask=pred)
    ax.set_title(
        f"KM curve for {TARGET.lower()}"
        f" and marker {marker} \n"
        f"(cutoff{direction}{cutoff:.2f}, log-rank-test p={res.p_value:.3f})"
    )
    ax.legend(
        [
            f"KP pred=0 (N={(~pred).sum()})",
            "95% CI (pred=0)",
            f"KP pred=1 (N={pred.sum()})",
            "95% CI (pred=1)",
        ]
    )
    fname = FOLDER / f"KM_plot_{marker}.pdf"
    files_out[fname.name] = fname
    njab.plotting.savefig(ax.get_figure(), fname)

    # add counts
    add_at_risk_counts(kmf_0, kmf_1, ax=ax)
    fname = FOLDER / f"KM_plot_{marker}_w_counts.pdf"
    files_out[fname.name] = fname
    njab.plotting.savefig(ax.get_figure(), fname)

# %%



"""
Lifeline plots. Adapted from
https://allendowney.github.io/SurvivalAnalysisPython/02_kaplan_meier.html
"""
import matplotlib
import matplotlib.pyplot as plt
import pandas as pd


def plot_lifelines(obs: pd.DataFrame,
                   ax=None,
                   start_col='DateDiagnose',
                   end_col='DateDeath',
                   status_col='dead') -> matplotlib.axes.Axes:
    """Plot a lifeline for each observation.
    """
    if ax is None:
        _, ax = plt.subplots()
    for i, (_, row) in enumerate(obs.iterrows()):
        start = row[start_col]
        end = row[end_col]
        status = row[status_col]

        if not status:
            # ongoing
            ax.hlines(i, start, end, color='C2')
        else:
            # complete
            ax.hlines(i, start, end, color='C1')
            ax.plot(end, i, marker='o', color='C1')
    return ax



from dataclasses import dataclass
from collections import namedtuple

import pandas as pd
import sklearn

AucRocCurve = namedtuple("AucRocCurve", 'fpr tpr cutoffs')
PrecisionRecallCurve = namedtuple("PrecisionRecallCurve",
                                  'precision recall cutoffs')


@dataclass
class ResultsSplit:
    """Results for a single split of train and test data."""
    auc: float = None  # receiver operation curve area under the curve
    aps: float = None  # average precision score
    roc: AucRocCurve = None
    prc: PrecisionRecallCurve = None


@dataclass
class Results:
    """Results for a model trained on a single split of train and test data."""
    model: sklearn.base.BaseEstimator = None
    selected_features: list = None
    train: ResultsSplit = None
    test: ResultsSplit = None
    name: str = None


@dataclass
class Splits:
    """Splits of train and test data with targets."""
    X_train: pd.DataFrame
    X_test: pd.DataFrame
    y_train: pd.Series
    y_test: pd.Series



import numpy as np
import pandas as pd
import sklearn.metrics as sklm

from njab.sklearn.types import Results


class ConfusionMatrix():
    """Wrapper for `sklearn.metrics.confusion_matrix`"""

    def __init__(self, y_true, y_pred):
        self.cm_ = sklm.confusion_matrix(y_true, y_pred)

    def as_dataframe(self, names=('true', 'pred')) -> pd.DataFrame:
        """Create pandas.DataFrame and return.
        Names rows and columns."""
        if not hasattr(self, 'df'):
            true_name, pred_name = names
            self.df = pd.DataFrame(self.cm_)
            self.df.index.name = true_name
            self.df.columns = pd.MultiIndex.from_product([[pred_name],
                                                          self.df.columns])
        return self.df

    def classification_label(self) -> dict:
        """Classification labels as dict."""
        tn, fp, fn, tp = self.cm_.ravel()
        return {'TN': tn, 'FP': fp, 'FN': fn, 'TP': tp}

    def as_classification_series(self) -> pd.Series:
        """Classification labels as pandas.Series."""
        return pd.Series(self.classification_label())

    @property
    def as_array(self):
        """Return sklearn.metrics.confusion_matrix array"""
        return self.cm_

    def __str__(self):
        """sklearn.metrics.confusion_matrix __str__"""
        return str(self.cm_)

    def __repr__(self):
        """sklearn.metrics.confusion_matrix __repr__"""
        return repr(self.cm_)


def get_label_binary_classification(y_true: int, y_pred: int) -> str:
    """Get labels (TP, FN, TN, FP) for single case in binary classification."""
    if y_true == 1:
        if y_pred == 1:
            return 'TP'
        elif y_pred == 0:
            return 'FN'
        else:
            ValueError(f"Unknown `y_pred`: {y_pred} ({ type(y_pred) })")
    elif y_true == 0:
        if y_pred == 0:
            return 'TN'
        elif y_pred == 1:
            return 'FP'
        else:
            ValueError(f"Unknown `y_pred`: {y_pred} ({ type(y_pred) })")
    else:
        raise ValueError(f"Unknown `y_true`: {y_true} ({ type(y_pred) })")


def get_score(clf, X: pd.DataFrame, pos=1) -> pd.Series:
    """Extract score from binary classifier for class one (target class)."""
    scores = clf.predict_proba(X)
    if scores.shape[-1] > 2:
        raise NotImplementedError
    else:
        scores = scores[:, pos]
    scores = pd.Series(scores, index=X.index)
    return scores


def get_pred(clf, X: pd.DataFrame) -> pd.Series:
    """Predict class for binary classifier and keep indices of from data X."""
    ret = clf.predict(X)
    ret = pd.Series(ret, index=X.index)
    return ret


def get_custom_pred(clf, X: pd.DataFrame, cutoff=0.5) -> pd.Series:
    """Calculate predicted class for binary classifier using the specified cutoff.
       Keep indices of from data X.
    """
    scores = get_score(clf, X)
    ret = (scores > cutoff).astype(int)
    return ret


def get_target_count_per_bin(score: pd.Series,
                             y: pd.Series,
                             n_bins: int = 10) -> pd.DataFrame:
    """Created pivot table with y summed per equality sized bin of scores."""
    pred_bins = pd.DataFrame({
        'score':
        pd.cut(score, bins=list(x / n_bins for x in range(0, n_bins + 1))),
        'y==1':
        y
    })
    pred_bins = pred_bins.groupby(by='score').sum().astype(int)
    return pred_bins


def get_lr_multiplicative_decomposition(results: Results, X: pd.DataFrame,
                                        prob: pd.Series,
                                        y: pd.Series) -> pd.DataFrame:
    """Multiplicative decompositon of odds at the base of the
    logistic regresion model."""
    components = X[results.selected_features].multiply(results.model.coef_)
    components['intercept'] = float(results.model.intercept_)
    components = np.exp(components)
    components['odds'] = prob / (1.0 - prob)
    components['prob'] = prob
    components[y.name] = y
    components = components.sort_values('prob', ascending=False)
    return components



try:
    import lifelines.statistics
    from lifelines import KaplanMeierFitter
except ModuleNotFoundError as e:
    msg = (
        "lifelines not available. Please install all njab dependencies typing:"
        "\n\tpip install 'njab[all]'"
        "\nor separately typing:\n\tpip install lifelines")
    e.args = (msg, *e.args[1:])
    raise ModuleNotFoundError(*e.args) from e
import pandas as pd
from matplotlib.axes import Axes


def compare_km_curves(
    time: pd.Series,
    y: pd.Series,
    pred: pd.Series,
    ax: Axes = None,
    ylim: tuple[int] = (0, 1),
    xlim: tuple[int] = (0, 180),
    xlabel: str = None,
    ylabel: str = None,
    add_risk_counts=False,
) -> Axes:
    """Compare Kaplan-Meier curves for two groups (e.g. based on binary prediction)

    Parameters
    ----------
    time : pd.Series
        Time to event variable
    y : pd.Series
        event variable
    pred : pd.Series
        mask for two groups, e.g. predictions
    ax : Axes, optional
        matplotlib Axes object, by default None
    ylim : tuple[int], optional
        y-axis bounds, by default (0, 1)
    xlim : tuple[int], optional
        time-axis bounds, by default (0, 730)
    xlabel : str, optional
        time-axis label, by default None
    ylabel : str, optional
        y-axis label, by default None


    Returns
    -------
    Axes, KaplanMeierFitter, KaplanMeierFitter
        Axes object,
        KaplanMeierFitter for predited as 0,
        KaplanMeierFitter for predicted as 1

    """
    pred = pred.astype(bool)
    mask = ~pred

    kmf_0 = KaplanMeierFitter()
    kmf_0.fit(time.loc[mask], event_observed=y.loc[mask])
    ax = kmf_0.plot(xlim=xlim, ylim=ylim, legend=False, ax=ax)

    mask = pred
    kmf_1 = KaplanMeierFitter()
    kmf_1.fit(time.loc[mask], event_observed=y.loc[mask])
    ax = kmf_1.plot(
        xlim=xlim,
        ylim=ylim,
        xlabel=xlabel,
        ylabel=ylabel,
        legend=False,
    )
    if add_risk_counts:
        from lifelines.plotting import add_at_risk_counts

        add_at_risk_counts(kmf_0, kmf_1, ax=ax)
    return ax, kmf_0, kmf_1


# ! move to stats:


def log_rank_test(
    time: pd.Series,
    y: pd.Series,
    mask: pd.Series,
) -> lifelines.statistics.StatisticalResult:
    """Compare Kaplan-Meier curves for two groups (e.g. based on binary prediction).
    Given the grouping in the mask, are both KM curves significantly different?

    Parameters
    ----------
    time : pd.Series
        Time to event variable
    y : pd.Series
        event variable
    mask : pd.Series
        mask for two groups, e.g. predictions


    Returns
    -------
    StatisticalResult:
        object containing results with a wrapper for visualization
    """
    mask = mask.astype(bool)
    res = lifelines.statistics.logrank_test(
        time.loc[mask],
        time.loc[~mask],
        y.loc[mask],
        y.loc[~mask],
    )
    return res
